#ifndef CTCP_H
#define CTCP_H

#include <qstringlist.h>
#include <qmap.h>
#include <qobject.h>

#include "singleton.h"

namespace IRC
{

class Window;
class Message;

class Message;

class CTCPMessage
{
public:
    enum QuotingFlag { Quoted, Unquoted };

    explicit CTCPMessage( const QString &msgType )
        : m_type( msgType ) {}
    CTCPMessage() {}

    bool isValid() const;

    const QString &type() const { return m_type; }

    static bool isCTCPMessage( const QString &text );

    static CTCPMessage fromString( const QString &text );

protected:
    QStringList parameters( QuotingFlag quoting = Quoted ) const;

    QString parameter( uint index, QuotingFlag quoting = Quoted ) const;

private:
    static QString dequoteString( const QString &input, QuotingFlag quoting );

    QString m_type;
    QStringList m_parameters;

    enum { CTCPMark = 0x01, SpaceQuote = '@' };
};

inline bool CTCPMessage::isValid() const
{
    return !m_type.isEmpty();
}

class CTCPAction : public CTCPMessage
{
public:
    CTCPAction( const CTCPMessage &body )
        : CTCPMessage( body ) {}
    CTCPAction() : CTCPMessage( name() ) {}

    QString rawText() const { return parameters( Unquoted ).join( " " ); }

    QString text( const Message &context ) const;

    static const char *name() { return "ACTION"; }
};

class CTCPHandler : public QObject
{
    Q_OBJECT
public:
    CTCPHandler( QObject *parent, const char *name = 0 );

public slots:
    void handleCTCPMessage( const QString &input, const IRC::Message &context, IRC::Window *window );

public:
    virtual void handleCTCPAction( const CTCPAction &msg, const IRC::Message &context, IRC::Window *window );
};

class CTCPMessageFactory : public Singleton<CTCPMessageFactory>
{
public:
    typedef CTCPMessage *( *CTCPMessageFactoryFunction )( const CTCPMessage &body );

    bool registerMessageType( const QString &typeName,
                              CTCPMessageFactoryFunction factoryFunc );

    template <class MessageImpl>
    bool registerMessageImpl();

    CTCPMessage *createMessage( const QString &typeName, const CTCPMessage &body );

private:
    typedef QMap<QString, CTCPMessageFactoryFunction> FactoryMap;

    FactoryMap m_typeMap;
};

template <class MessageImpl>
bool CTCPMessageFactory::registerMessageImpl()
{
    struct Factory
    {
        static CTCPMessage *create( const CTCPMessage &body ) { return new MessageImpl( body ); }
    };

    registerMessageType( MessageImpl::name(), Factory::create );
}

} //namespace IRC

#endif // CTCP_H
/* vim: et sw=4 ts=4
 */
