/* This file is part of the KDE project
   Copyright (C) 2002 Simon Hausmann <hausmann@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
#ifndef __message_h__
#define __message_h__

#include <qstringlist.h>

#include "global.h"
#include "reply.h"

namespace IRC
{

// forward decl.
// ### bad name
template <typename T, typename _SizeType = size_t> class BoundPointer;

class Message
{
public:
    typedef QMemArray<uint> NumericMessageTypeArray;
    struct NumericReplyInfo
    {
        NumericReplyInfo() {}
        NumericReplyInfo( const NumericMessageTypeArray &_replyCodes,
                          const NumericMessageTypeArray &_errorCodes )
            : replyCodes( _replyCodes ), errorCodes( _errorCodes )
        {}

        NumericMessageTypeArray replyCodes;
        NumericMessageTypeArray errorCodes;
    };

    Message();
    explicit Message( const QCString &type );
    Message( const QString &prefix, const QString &type, const QStringList &parameters );
    Message( const Message &rhs );
    Message &operator=( const Message &rhs );
    virtual ~Message();

    bool isValid() const;

    QString type() const { return m_type; }
    bool isNumericType() const;
    uint numericType() const;

    QString serialize() const;

    QString prefix() const { return m_prefix; }

    Person sender() const { return Person::fromMaskString( m_prefix ); }

    const QStringList &parameters() const { return m_parameters; }

    QString parameter( uint index ) const;

    static Message fromString( const QString &input );

protected:
    void setPrefix( const QString &prefix );
    void setType( const QString &type );
    void setParameters( const QStringList &parameters );

    void setParameter( uint index, const QString &value );

private:
    bool checkValidity() const;

    QString m_prefix;
    QString m_type;
    QStringList m_parameters;
    mutable bool m_validityChecked : 1;
    mutable bool m_cachedValidity : 1;

    QString serializePrefix() const;
    QString serializeType() const;
    QString serializeParameters() const;
    static QString serializeParameter( const QString &param, bool isLast );

    typedef BoundPointer<QChar> ParsingState;

    bool parsePrefix( ParsingState &state );
    bool parseType( ParsingState &state );
    bool parseParams( ParsingState &state );
    void parseSpace( ParsingState &state );
};

typedef QValueList<Message> MessageList;

inline Message::Message() : m_validityChecked( false ), m_cachedValidity( false ) {}

inline Message::Message( const QCString &type )
    : m_type( QString::fromLatin1( type ) ), m_validityChecked( true ), 
      m_cachedValidity( false )
{}

inline Message::Message( const QString &prefix, const QString &type,
                         const QStringList &parameters )
    : m_prefix( prefix ), m_type( type ), m_parameters( parameters ),
      m_validityChecked( true ), m_cachedValidity( true )
{}

inline Message::Message( const Message &rhs )
{ (*this) = rhs; }

inline Message &Message::operator=( const Message &rhs )
{
    m_prefix = rhs.m_prefix;
    m_type = rhs.m_type;
    m_parameters = rhs.m_parameters;
    m_validityChecked = rhs.m_validityChecked;
    m_cachedValidity = rhs.m_cachedValidity;
    return *this;
}

inline void Message::setPrefix( const QString &prefix )
{ m_prefix = prefix; m_validityChecked = false; }

inline void Message::setType( const QString &type )
{ m_type = type; m_validityChecked = false; }

inline void Message::setParameters( const QStringList &parameters )
{ m_parameters = parameters; m_validityChecked = false; }

class PassMessage : public Message
{
public:
    PassMessage() : Message( name() ) {}
    explicit PassMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "PASS"; }
    static NumericReplyInfo replyInfo();

    void setPassword( const QString &password ) { setParameter( 0, password ); }
    QString password() const { return parameter( 0 ); }
};

class NickMessage : public Message
{
public:
    NickMessage() : Message( name() ) {}
    explicit NickMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "NICK"; }
    static NumericReplyInfo replyInfo();

    void setNick( const QString &nick ) { setParameter( 0, nick ); }
    QString nick() const { return parameter( 0 ); }

    QString description() const;
};

class UserMessage : public Message
{
public:
    UserMessage() : Message( name() ) {}
    explicit UserMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "USER"; }
    static NumericReplyInfo replyInfo();

    void setUser( const QString &name ) { setParameter( 0, name ); }
    QString user() const { return parameter( 0 ); }
    void setHost( const QString &host ) { setParameter( 1, host ); }
    QString host() const { return parameter( 1 ); }
    void setServer( const QString &name ) { setParameter( 2, name ); }
    QString server() const { return parameter( 2 ); }
    void setRealName( const QString &name ) { setParameter( 3, name ); }
    QString realName() const { return parameter( 3 ); }
};

class ServerMessage : public Message
{
public:
    ServerMessage() : Message( name() ) {}
    explicit ServerMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "SERVER"; }
    static NumericReplyInfo replyInfo();
};

class OperMessage : public Message
{
public:
    OperMessage() : Message( name() ) {}
    explicit OperMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "OPER"; }
    static NumericReplyInfo replyInfo();
};

class QuitMessage : public Message
{
public:
    QuitMessage() : Message( name() ) {}
    explicit QuitMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "QUIT"; }
    static NumericReplyInfo replyInfo() { return NumericReplyInfo(); }

    void setQuitMessage( const QString &text ) { setParameter( 0, text ); }
    QString quitMessage() const { return parameter( 0 ); }

    QString description() const;
};

class SquitMessage : public Message
{
public:
    SquitMessage() : Message( name() ) {}
    explicit SquitMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "SQUIT"; }
    static NumericReplyInfo replyInfo();
};

class JoinMessage : public Message
{
public:
    JoinMessage() : Message( name() ) {}
    explicit JoinMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "JOIN"; }
    static NumericReplyInfo replyInfo();

    struct ChannelInfo
    {
        ChannelInfo() {}
        ChannelInfo( const QString &_channel, const QString &_key = QString::null )
            : channel( _channel ), key( _key ) {}

        QString channel;
        QString key;
    };

    class ChannelInfoList : public QValueList<ChannelInfo>
    {
    public:
        ChannelInfoList() : QValueList<ChannelInfo>() {}
        ChannelInfoList( const ChannelInfoList &rhs ) : QValueList<ChannelInfo>( rhs ) {}
        ChannelInfoList &operator=( const ChannelInfoList &rhs )
        { QValueList<ChannelInfo>::operator=( rhs ); return *this; }

        QStringList channelList() const
        {
            QStringList result;
            for ( ConstIterator it = begin(); it != end(); ++it )
                result << ( *it ).channel;
            return result;
        }
    };

    ChannelInfoList channelInfo() const;
    void setChannelInfo( const ChannelInfoList &channels );

    QString description( const QString &channel ) const;
};

class PartMessage : public Message
{
public:
    PartMessage() : Message( name() ) {}
    explicit PartMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "PART"; }
    static NumericReplyInfo replyInfo();

    void setChannel( const QString &channel )
    { setParameter( 0, channel ); }

    void setChannels( const QStringList &channels )
    { setParameter( 0, channels.join( "," ) ); }

    QStringList channels() const 
    { return QStringList::split( ",", parameter( 0 ) ); }

    QString reason() const
    { return parameter( 1 ); }

    void setReason( const QString &reason )
    { setParameter( 1, reason ); }

    QString description( const QString &channel ) const;
};

class ModeMessage : public Message
{
public:
    ModeMessage() : Message( name() ) {}
    explicit ModeMessage( const Message &body ) : Message( body ) {}

    void setChannel( const QString &channel ) { setParameter( 0, channel ); }
    QString channel() const { return parameter( 0 ); }

    static const char *name() { return "MODE"; }
    static NumericReplyInfo replyInfo() { return NumericReplyInfo(); }
};

class TopicMessage : public Message
{
public:
    TopicMessage() : Message( name() ) {}
    explicit TopicMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "TOPIC"; }
    static NumericReplyInfo replyInfo();

    void setChannel( const QString &channel ) { setParameter( 0, channel ); }
    QString channel() const { return parameter( 0 ).lower(); }

    void setTopic( const QString &text ) { setParameter( 1, text ); }
    QString topic() const { return parameter( 1 ); }
};

class NamesMessage : public Message
{
public:
    NamesMessage() : Message( name() ) {}
    explicit NamesMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "NAMES"; }
    static NumericReplyInfo replyInfo();

    void setChannels( const QStringList &channels )
    { setParameter( 0, channels.join( "," ) ); }

    QStringList channels() const 
    { return QStringList::split( ",", parameter( 0 ) ); }
};

class ListMessage : public Message
{
public:
    ListMessage() : Message( name() ) {}
    explicit ListMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "LIST"; }
    static NumericReplyInfo replyInfo();
};

class InviteMessage : public Message
{
public:
    InviteMessage() : Message( name() ) {}
    explicit InviteMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "INVITE"; }
    static NumericReplyInfo replyInfo();
};

class KickMessage : public Message
{
public:
    KickMessage() : Message( name() ) {}
    explicit KickMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "KICK"; }
    static NumericReplyInfo replyInfo();
};

class VersionMessage : public Message
{
public:
    VersionMessage() : Message( name() ) {}
    explicit VersionMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "VERSION"; }
    static NumericReplyInfo replyInfo();
};

class StatsMessage : public Message
{
public:
    StatsMessage() : Message( name() ) {}
    explicit StatsMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "STATS"; }
    static NumericReplyInfo replyInfo();
};

class LinksMessage : public Message
{
public:
    LinksMessage() : Message( name() ) {}
    explicit LinksMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "LINKS"; }
    static NumericReplyInfo replyInfo();
};

class TimeMessage : public Message
{
public:
    TimeMessage() : Message( name() ) {}
    explicit TimeMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "TIME"; }
    static NumericReplyInfo replyInfo();
};

class ConnectMessage : public Message
{
public:
    ConnectMessage() : Message( name() ) {}
    explicit ConnectMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "CONNECT"; }
    static NumericReplyInfo replyInfo();
};

class TraceMessage : public Message
{
public:
    TraceMessage() : Message( name() ) {}
    explicit TraceMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "TRACE"; }
    static NumericReplyInfo replyInfo();
};

class AdminMessage : public Message
{
public:
    AdminMessage() : Message( name() ) {}
    explicit AdminMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "ADMIN"; }
    static NumericReplyInfo replyInfo();
};

class InfoMessage : public Message
{
public:
    InfoMessage() : Message( name() ) {}
    explicit InfoMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "INFO"; }
    static NumericReplyInfo replyInfo();
};

class PrivmsgMessage : public Message
{
public:
    PrivmsgMessage() : Message( name() ) {}
    explicit PrivmsgMessage( const Message &body ) : Message( body ) {}
    explicit PrivmsgMessage( const char *name ) : Message( name ) {}

    static const char *name() { return "PRIVMSG"; }
    static NumericReplyInfo replyInfo();

    void setReceiver( const QString &receiver )
    { setParameter( 0, receiver ); }
    void setReceivers( const QStringList &receivers ) 
    { setParameter( 0, receivers.join( "," ) ); }
    QStringList receivers() const { return QStringList::split( ",", parameter( 0 ) ); }

    void setPlainText( const QString &text ) { setParameter( 1, text ); }
    QString plainText() const { return parameter( 1 ); }

    QString text() const;

    void setAction( const QString &actionText );

    bool isCTCPMessage() const;
};

class NoticeMessage : public PrivmsgMessage
{
public:
    NoticeMessage() : PrivmsgMessage( name() ) {}
    explicit NoticeMessage( const Message &body ) : PrivmsgMessage( body ) {}

    static const char *name() { return "NOTICE"; }
    static NumericReplyInfo replyInfo() { return NumericReplyInfo(); }
};

class WhoMessage : public Message
{
public:
    WhoMessage() : Message( name() ) {}
    explicit WhoMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "WHO"; }
    static NumericReplyInfo replyInfo();
};

class WhoisMessage : public Message
{
public:
    WhoisMessage() : Message( name() ) {}
    explicit WhoisMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "WHOIS"; }
    static NumericReplyInfo replyInfo();

    void setTarget( const QString &nick )
    { setParameter( 0, nick ); }

    QString target() const
    { return parameter( 0 ); }
};

class WhowasMessage : public Message
{
public:
    WhowasMessage() : Message( name() ) {}
    explicit WhowasMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "WHOWAS"; }
    static NumericReplyInfo replyInfo();
};

class KillMessage : public Message
{
public:
    KillMessage() : Message( name() ) {}
    explicit KillMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "KILL"; }
    static NumericReplyInfo replyInfo();
};

class PongMessage : public Message
{
public:
    PongMessage() : Message( name() ) {}
    explicit PongMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "PONG"; }
    static NumericReplyInfo replyInfo();

    void setDaemons( const QStringList &daemons )
    { setParameters( daemons ); }
};

class PingMessage : public Message
{
public:
    PingMessage() : Message( name() ) {}
    explicit PingMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "PING"; }
    static NumericReplyInfo replyInfo();

    void setServers( const QStringList &servers )
    { setParameters( servers ); }

    QStringList servers() const { return parameters(); }

    Message pong() const 
    {
        PongMessage reply;
        reply.setDaemons( servers() );
        return reply;
    }
};

class ErrorMessage : public Message
{
public:
    ErrorMessage() : Message( name() ) {}
    explicit ErrorMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "ERROR"; }
    static NumericReplyInfo replyInfo() { return NumericReplyInfo(); }
};

class AwayMessage : public Message
{
public:
    AwayMessage() : Message( name() ) {}
    explicit AwayMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "AWAY"; }
    static NumericReplyInfo replyInfo();
};

class RehashMessage : public Message
{
public:
    RehashMessage() : Message( name() ) {}
    explicit RehashMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "REHASH"; }
    static NumericReplyInfo replyInfo();
};

class RestartMessage : public Message
{
public:
    RestartMessage() : Message( name() ) {}
    explicit RestartMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "RESTART"; }
    static NumericReplyInfo replyInfo();
};

class SummonMessage : public Message
{
public:
    SummonMessage() : Message( name() ) {}
    explicit SummonMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "SUMMON"; }
    static NumericReplyInfo replyInfo();
};

class UsersMessage : public Message
{
public:
    UsersMessage() : Message( name() ) {}
    explicit UsersMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "USERS"; }
    static NumericReplyInfo replyInfo();
};

/* server side only
class WallopsMessage : public Message
{
public:
    WallopsMessage() : Message( name() ) {}
    explicit WallopsMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "WALLOPS"; }
    typedef K_CONSTLIST_1( ERR_NEEDMOREPARAMS ) ErrorCodes;
};
*/

class UserhostMessage : public Message
{
public:
    UserhostMessage() : Message( name() ) {}
    explicit UserhostMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "USERHOST"; }
    static NumericReplyInfo replyInfo();
};

class IsonMessage : public Message
{
public:
    IsonMessage() : Message( name() ) {}
    explicit IsonMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "ISON"; }
    static NumericReplyInfo replyInfo();
};

// ### wrong. should be a simple replyhandler of a IRCServer class or the like
/*
class ServerInfoMessage : public Message
{
public:
    ServerInfoMessage() : Message( name() ) {}
    explicit ServerInfoMessage( const Message &body ) : Message( body ) {}

    static const char *name() { return "004"; }
    typedef KDE::NullType ErrorCodes;

    QString version() const { return parameter( 2 ); }
    QString userModes() const { return parameter( 3 ); }
    QString channelModes() const { return parameter( 4 ); }
};
*/

};

#endif

/* vim: et sw=4
 */
